@@ -94,7 +94,9 @@ class MeshNode (GeometryNode):
         if self.resource is not None and len(self.resource.triangles)>=2:
             pass
         elif self.get_path():
-            self._resource   =  o3d.io.read_triangle_mesh(self.path)
+            resource =  o3d.io.read_triangle_mesh(self.path)
+            if len(resource.triangles)>2:
+                self._resource  =resource
         return self._resource  
 
     def save_resource(self, directory:str=None,extension :str = '.ply') ->bool:
