@@ -347,10 +347,15 @@ class SessionNode(GeometryNode):
         if self.resource is None:
             return False
         
+        #validate extension
+        if extension not in ut.MESH_EXTENSION:
+            raise ValueError('Invalid error')
+
         # check if already exists
         if directory and os.path.exists(os.path.join(directory,self.get_name() + extension)):
+            self.path=os.path.join(directory,self.get_name() + extension)
             return True
-        elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.MESH_EXTENSION:
+        elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.MESH_EXTENSION:            
             return True
                     
         #get directory
@@ -367,11 +372,8 @@ class SessionNode(GeometryNode):
         if not os.path.exists(directory):                        
             os.mkdir(directory) 
 
-        #check extension
-        if extension in ut.MESH_EXTENSION:
-            self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)
+        self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)
 
-        
         #write files
         if o3d.io.write_triangle_mesh(self.path, self.resource):
             return True
