@@ -543,7 +543,8 @@ def graph_path_to_nodes(graphPath : str,**kwargs) -> List[Node]:
         nodelist=[]
         graph=Graph().parse(graphPath)
         for subject in graph.subjects(RDF.type):
-            nodelist.append(create_node(graph=graph,graphPath=graphPath,subject=subject,**kwargs) )
+            myGraph=ut.get_subject_graph(graph,subject)
+            nodelist.append(create_node(graph=myGraph,graphPath=graphPath,subject=subject,**kwargs) )
         return nodelist
     else:
         raise ValueError('No valid graphPath (only .ttl).')
@@ -583,6 +584,52 @@ def graph_to_nodes(graph : Graph,**kwargs) -> List[Node]:
 #     else:
 #         node=Node(graph=g,**kwargs) 
 #     return node
+def create_node(graph: Graph = None, graphPath: str =None, subject: URIRef = None, resource = None, **kwargs)-> Node:
+    """_summary_
+
+    Args:
+        graph (Graph, optional): _description_. Defaults to None.
+        graphPath (str, optional): _description_. Defaults to None.
+        subject (URIRef, optional): _description_. Defaults to None.
+
+    Returns:
+        Node (PointCloudNode,MeshNode,GeometryNode,ImageNode)
+    """
+    #input validation
+    if(graphPath and not graph):
+            graph = Graph().parse(graphPath)
+    if(graph and not subject):
+        subject=next(graph.subjects(RDF.type))
+    if (subject and graph):    
+        nodeType = ut.literal_to_string(graph.value(subject=subject,predicate=RDF.type))
+    elif (resource):
+        if type(resource) is o3d.geometry.PointCloud:
+            nodeType='PointCloudNode'
+        elif type(resource) is o3d.geometry.TriangleMesh:
+            nodeType='MeshNode'
+        elif type(resource) is o3d.geometry:
+            nodeType='GeometryNode'
+        elif type(resource) is np.ndarray:
+            nodeType='ImageNode'        
+    else:        
+        nodeType = 'Node'
+
+    #node creation
+    if 'BIMNode' in nodeType:
+        node=BIMNode(graph=graph, graphPath=graphPath, resource=resource,subject=subject, **kwargs)
+    elif 'MeshNode' in nodeType:
+        node=MeshNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
+    elif 'GeometryNode' in nodeType:
+        node=GeometryNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
+    elif 'PointCloudNode' in nodeType:
+        node=PointCloudNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
+    elif 'ImageNode' in nodeType:
+        node=ImageNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
+    elif 'SessionNode' in nodeType:
+        node=SessionNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)  
+    else:
+        node=Node(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs) 
+    return node
 
 def get_linked_nodes(node: Node ,graph:Graph, getResource=False, **kwargs) -> List[Node]:
     """Get related nodes based on linkedNodes variable.\n
