@@ -42,10 +42,35 @@ ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')
 #### BASIC OPERATIONS ####
 
 def get_extension(path:str) -> str:
+    """Returns the filepath extensions.
+    \n
+    E.g. D://myData//test.txt -> .txt
+
+    Args:
+        path (str):
+
+    Returns:
+        str: 
+    """
     filename,extension = os.path.splitext(path)
     return extension
 
-def replace_str_index(text,index=0,replacement='_'):
+def replace_str_index(text:str,index:int=0,replacement:str='_'):
+    """Replace a string character at the location of the index with the replacement.
+    \n
+
+    replacement=_\n
+    index=1\n
+    text= 'text' -> t_xt\n
+
+    Args:
+        0. text (str)\n
+        index (int, optional): _description_. Defaults to 0.
+        replacement (str, optional): _description_. Defaults to '_'.
+
+    Returns:
+        _type_: _description_
+    """
     return '%s%s%s'%(text[:index],replacement,text[index+1:])
 
 
