@@ -15,7 +15,7 @@ from PIL import Image
 
 import xml.etree.ElementTree as ET 
 import math
-from typing import List
+from typing import List,Tuple
 import matplotlib
 import os
 
@@ -122,15 +122,15 @@ def mesh_to_trimesh(geometry: o3d.geometry) -> trimesh.Trimesh:
         return None
 
 def crop_geometry_by_convex_hull(source:trimesh.Trimesh, cutters: List[trimesh.Trimesh], inside : Boolean = True ) -> trimesh.Trimesh:
-    """"Cut a portion of a mesh that lies within the convex hull of another mesh.
+    """"Cut a portion of a mesh that lies within the convex hull of another mesh.\n
     
-    .. image:: ../pics/crop_by_convex_hull.PNG
+    .. image:: ../../docs/pics/crop_by_convex_hull.PNG
 
     Args:
-        source (trimesh.Trimesh):   mesh that will be cut
-        cutter (trimesh.Trimesh):   mesh of which the faces are used for the cuts. Face normals should point outwards (positive side)
-        strict (Boolean):           True if source faces can only be part of a single submesh
-        inside (Boolean):           True if retain the inside of the intersection
+        1. source (trimesh.Trimesh):   mesh that will be cut \n
+        2. cutter (trimesh.Trimesh):   mesh of which the faces are used for the cuts. Face normals should point outwards (positive side) \n
+        3. strict (Boolean):           True if source faces can only be part of a single submesh\n
+        4. inside (Boolean):           True if retain the inside of the intersection\n
     Returns:
         mesh:       trimesh.Trimesh 
         None:       if no data was retained
@@ -200,12 +200,12 @@ def get_mesh_representation(geometry)->o3d.geometry.TriangleMesh:
         raise ValueError('No valid input: o3d.geometry.TriangleMesh or o3d.geometry.PointCloud')
 
 def get_mesh_collisions_trimesh(sourceMesh: o3d.geometry.TriangleMesh, geometries: List[o3d.geometry.TriangleMesh]) -> List[int]:
-    """Return indices of geometries that collide with the source.
+    """Return indices of geometries that collide with the source.\n
 
-    .. image:: ../pics/collision_4.PNG
+    .. image:: ../../docs/pics/collision_4.PNG
 
     Args:
-        sourceMesh (o3d.geometry.TriangleMesh)
+        sourceMesh (o3d.geometry.TriangleMesh)\n
         geometries (List[o3d.geometry.TriangleMesh])
 
     Returns:
@@ -536,17 +536,17 @@ def mesh_to_pcd(mesh:o3d.geometry.TriangleMesh,voxelSize : float = 0.1) -> o3d.g
 #### GEOMETRY GENERATION ####
 
 def generate_virtual_images(geometries: List[o3d.geometry.Geometry],cartesianTransforms: List[np.array],width:int=640,height:int=480,f:float=400)-> o3d.geometry.Image:
-    """Generate a set of Open3D Images from cartesianTransforms and geometries. 
-    The same intrinsic camera parameters are used for all cartesianTransforms that are passed to the function
+    """Generate a set of Open3D Images from cartesianTransforms and geometries. \n
+    The same intrinsic camera parameters are used for all cartesianTransforms that are passed to the function.\n
 
-    .. image:: ../docs/pics/rendering3.PNG
+    .. image:: ../../docs/pics/rendering3.PNG
 
     Args:
-        geometries (List[o3d.geometry]):o3d.geometry.PointCloud or o3d.geometry.TriangleMesh
-        cartesianTransforms (List[np.array 4x4]): [Rt]
-        width (int, optional): image width in pix. Defaults to 640.
-        height (int, optional): image height in pix. Defaults to 480.
-        f (float, optional): focal length in pix. Defaults to 400.
+        1. geometries (List[o3d.geometry]):o3d.geometry.PointCloud or o3d.geometry.TriangleMesh \n
+        2. cartesianTransforms (List[np.array 4x4]): [Rt] \n
+        3. width (int, optional): image width in pix. Defaults to 640pix.\n
+        4. height (int, optional): image height in pix. Defaults to 480pix. \n
+        5. f (float, optional): focal length in pix. Defaults to 400pix. \n
 
     Returns:
         List[o3d.cpu.pybind.geometry.Image]
@@ -708,14 +708,14 @@ def get_matching_indices(source, target, trans, search_voxel_size, K=None):
 #     return correspondences
 
 def crop_geometry_by_box(geometry:o3d.geometry, box:o3d.geometry.OrientedBoundingBox, subdivide:int = 0) ->o3d.geometry:
-    """"Crop portion of a mesh/pointcloud that lies within an OrientedBoundingBox 
+    """"Crop portion of a mesh/pointcloud that lies within an OrientedBoundingBox.\n
 
-    .. image:: ../pics/selection_BB_mesh2.PNG
+    .. image:: ../../docs/pics/selection_BB_mesh2.PNG
 
     Args:
-        geometry (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud)
-        box (o3d.geometry.OrientedBoundingBox )
-        subdivide (int): number of interations to increase the density of the mesh (1=x4, 2=x16, etc.)
+        1. geometry (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud): Geometry to be cropped\n
+        2. box (o3d.geometry.OrientedBoundingBox ): bouding cutter geometry\n
+        3. subdivide (int): number of interations to increase the density of the mesh (1=x4, 2=x16, etc.)\n
         
     Returns:
         o3d.geometry.TriangleMesh or None
@@ -784,24 +784,171 @@ def join_geometries(geometries) ->o3d.geometry:
         o3d.geometry: 
     """
     geometries=ut.item_to_list(geometries)
-    if any(isinstance(val, o3d.cpu.pybind.geometry.TriangleMesh) for val in geometries):
+    if any('TriangleMesh' in str(type(g)) for g in geometries ):
         joined=o3d.geometry.TriangleMesh()
         for geometry in geometries:
             if geometry is not None and len(geometry.vertices) != 0:
                 joined +=geometry
         return joined
-    elif any(isinstance(val, o3d.geometry.PointCloud) for val in geometries): 
+    if any('PointCloud' in str(type(g)) for g in geometries ):
         joined=o3d.geometry.PointCloud()
         for geometry in geometries:
             if geometry is not None and len(geometry.points) != 0:
                 joined +=geometry
         return joined
     else:
-        print('submit point clouds or meshes')
-        return None
+        raise ValueError('Only submit o3d.geometry.TriangleMesh or o3d.geometry.PointCloud objects') 
 
-def determine_percentage_of_coverage(sources: o3d.geometry, reference:o3d.geometry,threshold:float=0.1)-> np.array:
+def create_visible_point_cloud_from_meshes (geometries: List[o3d.geometry.TriangleMesh], references:List[o3d.geometry.TriangleMesh], resolution:float = 0.1)-> Tuple[List[o3d.geometry.PointCloud], List[float]]:
+    """Returns a set of point clouds sampled on the geometries.\n
+    Each point cloud has its points filtered to not lie wihtin or collide with any of the reference geometries. \n
+    As such, this method returns the **visible** parts of a set of sampled point clouds. \n
+    For every point cloud, the percentage of visibility is also reported. \n
+    This method takes about 50s for 1000 geometries. \n
 
+    E.g. The figure shows the points of the visible point cloud that were rejected due to their proximity to the other mesh geometries
+
+    .. image:: ../../docs/pics/invisible_points.PNG
+
+    Args:
+        geometries (List[o3d.geometry.TriangleMesh]): Meshes that will be sampled up to the resolution. \n
+        references (List[o3d.geometry.TriangleMesh]): reference meshes that are used to spatially filter the sampled point clouds so only 'visible' points are retained. \n
+        resolution (float, optional): Spatial resolution to sample meshes. Defaults to 0.1m. \n
+
+    Raises:
+        ValueError: any('TriangleMesh' not in str(type(g)) for g in geometries )\n
+        ValueError: any('TriangleMesh' not in str(type(g)) for g in references )\n
+
+    Returns:
+        Tuple[List[o3d.geometry.PointCloud], List[percentages [0-1.0]]]: 
+    """
+    geometries=ut.item_to_list(geometries)    
+    references=ut.item_to_list(references)    
+
+    #validate geometries
+    if  any('TriangleMesh' not in str(type(g)) for g in geometries ):
+        raise ValueError('Only submit o3d.geometry.TriangleMesh objects') 
+    #validate geometries
+    if  any('TriangleMesh' not in str(type(g)) for g in references ):
+        raise ValueError('Only submit o3d.geometry.TriangleMesh objects') 
+
+    colorArray=np.random.random((len(geometries),3))
+    identityPointClouds=[]
+    percentages=[]
+    # percentages=np.full((len(pcd.points), 1), 1.0)
+
+    for i,geometry in enumerate( geometries):
+        # create a reference scene 
+        referenceGeometries=[g for g in references if g !=geometry ]
+        reference=join_geometries(referenceGeometries)
+        scene = o3d.t.geometry.RaycastingScene()
+        cpuReference = o3d.t.geometry.TriangleMesh.from_legacy(reference)
+        _ = scene.add_triangles(cpuReference)
+
+        # sample mesh
+        area=geometry.get_surface_area()
+        count=int(area/(resolution*resolution))
+        pcd=geometry.sample_points_uniformly(number_of_points=10*count)
+        pcd=pcd.voxel_down_sample(resolution)
+
+        # determine visibility from distance and occupancy querries
+        query_points = o3d.core.Tensor(np.asarray(pcd.points), dtype=o3d.core.Dtype.Float32)
+        unsigned_distance = scene.compute_distance(query_points)
+        occupancy = scene.compute_occupancy(query_points)
+        indices=np.where((unsigned_distance.numpy() >=0.5*resolution) & (occupancy.numpy() ==0) )[0]     
+
+        # crop sampled point cloud to only the visible points
+        pcdCropped = pcd.select_by_index(indices)
+        pcdCropped.paint_uniform_color(colorArray[i])
+        identityPointClouds.append(pcdCropped)
+
+        #report percentage
+        percentages.append((len(pcdCropped.points)/len(pcd.points)))
+
+    return identityPointClouds, percentages
+
+def create_identity_point_cloud(geometries: o3d.geometry, resolution:float = 0.1) -> Tuple[o3d.geometry.PointCloud, np.array]:
+    """Returns a sampled point cloud colorized per object of a set of objects along with an array with an identifier for each point.
+
+    Args:
+        1.geometries (o3d.geometry.PointCloud or o3d.geometry.TriangleMesh) \n
+        2.resolution (float, optional): (down)sampling resolution for the point cloud. Defaults to 0.1.
+
+    Raises:
+        ValueError: Geometries must be o3d.geometry (PointCloud or TriangleMesh)
+
+    Returns:
+        Tuple[colorized point cloud (o3d.geometry.PointCloud), identityArray(np.array)]
+    """
+    geometries=ut.item_to_list(geometries)    
+    colorArray=np.random.random((len(geometries),3))
+    indentityArray=None
+    identityPointCloud=o3d.geometry.PointCloud()
+
+    for i,geometry in enumerate(geometries):
+        if 'PointCloud' in str(type(geometry)) :
+            pcd=geometry.voxel_down_sample(resolution)
+            indentityArray=np.vstack((indentityArray,np.full((len(pcd.points), 1), i)))
+            pcd.paint_uniform_color(colorArray[i])
+            identityPointCloud +=pcd
+            # np.concatenate((np.asarray(identityPointCloud.points),np.asarray(identityPointCloud.points)),axis=0)
+        elif 'TriangleMesh' in str(type(geometry)):
+            area=geometry.get_surface_area()
+            count=int(area/(resolution*resolution))
+            pcd=geometry.sample_points_uniformly(number_of_points=10*count)
+            pcd=pcd.voxel_down_sample(resolution)
+            indentityArray=np.vstack((indentityArray,np.full((len(pcd.points), 1), i)))
+            pcd.paint_uniform_color(colorArray[i])
+            identityPointCloud +=pcd
+        else:
+            print(f'{geometry} is invalid')
+            continue
+            # raise ValueError('Geometries must be o3d.geometry (PointCloud or TriangleMesh)')    
+    indentityArray=indentityArray.flatten()
+    indentityArray=np.delete(indentityArray,0)
+    return identityPointCloud, indentityArray
+
+def determine_observable_point_cloud(pointcloud:o3d.geometry.PointCloud, resolution:float = 0.1,indentityArray:np.ndarray=None )-> Tuple[o3d.geometry.PointCloud, np.array]:
+
+    #crop point cloud
+    dists = np.asarray(pointcloud.compute_point_cloud_distance(pointcloud))
+    indices = np.where((dists < resolution) and (dists >0.001))[0]
+    pcdCropped = pointcloud.select_by_index(indices)
+
+    if indentityArray:
+        #compute percentages based on occurences
+        unique0, counts0 = np.unique(indentityArray, return_counts=True)
+        percentages=np.full((len(unique0), 1), 1)
+        arrayCropped=indentityArray[list(indices)]
+        unique1, counts1 = np.unique(arrayCropped, return_counts=True)
+        for u in unique1:
+            percentages[u]=counts1[u]/counts0[u]
+    else:
+        percentages=None
+    return pcdCropped, percentages
+
+def determine_percentage_of_coverage(sources: List[o3d.geometry], reference:o3d.geometry.PointCloud,threshold:float=0.1)-> np.array:
+    """Returns the Percentage-of-Coverage (PoC) of every source geometry when compared to a reference geometry. \n
+    The PoC is defined as the ratio of points on the boundary surface of the source that lie within a Euclidean distance threshold hold of the reference geometry. \n
+    sampled point cloud on the boundary surface of the sources with a resolution of e.g. 0.1m \n
+
+    $$ P_{i'}=\{ p|\forall p \in P_i : p_i \cap N\backslash  n_i \} $$
+
+    $$ c_i = \frac{|P_{i'}|}{|P_i|}$$
+
+    E.g. a mesh of a beam of which half the surface lies within 0.1m of a point cloud will have a PoC of 0.5
+    
+    Args:
+        sources (o3d.geometry.TriangleMesh/PointCloud): geometries to determine the PoC for. \n
+        reference (o3d.geometry.PointCloud): reference geometry for the Euclidean distance calculations.\n
+        threshold (float, optional): sampling resolution of the boundary surface of the source geometries. Defaults to 0.1m.\n
+
+    Raises:
+        ValueError: Sources must be o3d.geometry (PointCloud or TriangleMesh)
+
+    Returns:
+        np.array (len(sources),1): percentages [0-1.0]
+    """
     #if no list, list
     sources=ut.item_to_list(sources)
     sourcePCDs=[]
@@ -851,14 +998,14 @@ def determine_percentage_of_coverage(sources: o3d.geometry, reference:o3d.geomet
     return percentages
 
 def crop_geometry_by_distance(source: o3d.geometry, cutters, threshold : float =0.10) -> o3d.geometry.PointCloud:
-    """"Select the portion of a pointcloud that lies within a range of another mesh/point cloud 
+    """"Returns the portion of a pointcloud that lies within a range of another mesh/point cloud.
     
-    .. image:: ../pics/crop_by_distance2.PNG
+    .. image:: ../../docs/pics/crop_by_distance2.PNG
 
     Args:
-        source (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud) : point cloud to filter
-        cutters (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud): list of reference data
-        threshold (float): threshold Euclidean distance for the filtering
+        1. source (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud) : point cloud to filter \n
+        2. cutters (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud): list of reference data \n
+        3. threshold (float, optional): threshold Euclidean distance for the filtering \n
 
     Returns:
         (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud)
@@ -911,16 +1058,15 @@ def crop_geometry_by_distance(source: o3d.geometry, cutters, threshold : float =
 #### GEOMETRY PROPERTIES
 
 def get_box_inliers(sourceBox:o3d.geometry.OrientedBoundingBox, testBoxes: List[o3d.geometry.OrientedBoundingBox])->List[int]:
-    """"
-    Return the indices of the testBoxes of which the bounding points lie within the sourceBox
+    """"Return the indices of the testBoxes of which the bounding points lie within the sourceBox.\n
 
-    .. image:: ../pics/get_box_inliers1.PNG
+    .. image:: ../../docs/pics/get_box_inliers1.PNG
 
     Args:
-        sourceBox (o3d.geometry.OrientedBoundingBox):   box to test
-        testBoxes (o3d.geometry.OrientedBoundingBox):   boxes to test
+        1. sourceBox (o3d.geometry.OrientedBoundingBox):   box to test \n
+        2. testBoxes (o3d.geometry.OrientedBoundingBox):   boxes to test \n
     Returns:
-        list (List[int]):       indices of testBoxes
+        list (List[int]):       indices of testBoxes \n
 
     """
     #Convert to list
@@ -936,9 +1082,22 @@ def get_box_inliers(sourceBox:o3d.geometry.OrientedBoundingBox, testBoxes: List[
     list = [ i for i in range(len(array)) if array[i]]
     return list
 
+def get_mesh_inliers(sources:List[o3d.geometry], reference:o3d.geometry) -> List[int]:
+    """Returns the indices of the geometries that lie within a reference geometry.\n
+       The watertightness of both source and reference geometries is determined after which the occupancy is computed for a numer of sampled points in the source geometries. \n
 
-def get_mesh_inliers(sources:o3d.geometry, reference:o3d.geometry) -> List[int]:
+    .. image:: ../../docs/pics/selection_BB_mesh5.PNG
+    
+    Args:
+        sources (o3d.geometry.TriangleMesh/PointCloud): geometries to test for collision
+        reference (o3d.geometry): reference geometry 
 
+    Raises:
+        ValueError: Only TriangleMesh and PointCloud allowed
+
+    Returns:
+        List[int]:indices of the inlier geometries
+    """
 
     sources=ut.item_to_list(sources)
     inliers=[None]*len(sources)
@@ -954,7 +1113,7 @@ def get_mesh_inliers(sources:o3d.geometry, reference:o3d.geometry) -> List[int]:
     elif ('PointCloud' in str(type(reference)) and len(reference.points)>=3):
         reference,_=reference.compute_convex_hull()
     else: 
-        raise ValueError(' Only TriangleMesh and PointCloud allowed')
+        raise ValueError('Only TriangleMesh and PointCloud allowed')
     #add reference to scene
     scene = o3d.t.geometry.RaycastingScene()
     reference = o3d.t.geometry.TriangleMesh.from_legacy(reference)
@@ -991,7 +1150,7 @@ def get_box_intersections(sourceBox:o3d.geometry.OrientedBoundingBox, testBoxes:
     2 oriented bounding boxes (A,B) overlap if the projection from B in the coordinate system of A on all the axes overlap. 
     The projection of B on the oriented axes of A is simply the coordinate range for that axis.
 
-    .. image:: ../pics/get_box_inliers1.PNG
+    .. image:: ../../docs/pics/get_box_inliers1.PNG
 
     Args:
         sourceBox (o3d.geometry.OrientedBoundingBox):   box to test
@@ -1233,19 +1392,19 @@ def get_mesh_collisions_open3d(sourceMesh: o3d.geometry.TriangleMesh, meshes: Li
         return False
 
 def crop_geometry_by_raycasting(source:o3d.geometry.TriangleMesh, cutter: o3d.geometry.TriangleMesh, inside : Boolean = True,strict : Boolean = True ) -> o3d.geometry.TriangleMesh:
-    """"Select portion of a mesh that lies within a mesh shape (if not closed, convexhull is called)
+    """"Select portion of a mesh that lies within a mesh shape (if not closed, convexhull is called).\n
     
     inside=True
-    .. image:: ../pics/crop_by_ray_casting1.PNG
+    .. image:: ../../docs/pics/crop_by_ray_casting1.PNG
 
     inside=False 
-    .. image:: ../pics/crop_by_ray_casting2.PNG
+    .. image:: ../../docs/pics/crop_by_ray_casting2.PNG
 
     Args:
-        source (o3d.geometry.TriangleMesh) : mesh to cut
-        cutter (o3d.geometry.TriangleMesh) : mesh that cuts
-        inside (bool): 'True' to retain inside
-        strict (bool): 'True' if no face vertex is allowed outside the bounds, 'False' allows 1 vertex to lie outside
+        1. source (o3d.geometry.TriangleMesh) : mesh to cut \n
+        2. cutter (o3d.geometry.TriangleMesh) : mesh that cuts \n
+        3. inside (bool): 'True' to retain inside \n
+        4. strict (bool): 'True' if no face vertex is allowed outside the bounds, 'False' allows 1 vertex to lie outside \n
 
     Returns:
         outputmesh(o3d.geometry.TriangleMesh)
